#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include <string.h>
#include <glade/glade.h>

typedef struct
{
	GtkWidget *from;
	GtkWidget *to;
	GtkWidget *debug;
} Flow;

#define ENTRY \
	printf("Entering function: %s\n", __FUNCTION__)

void compile_file(gchar *c_text, gchar **asm_text, gchar **debug_output);


void on_menu_help_about_activate(GtkWidget *widget, gpointer user_data)
{
    gtk_widget_show(widget);
}

void on_about_dialog_response(GtkWidget *widget, gpointer user_data)
{
    gtk_widget_hide(widget);
}

void compile_button_clicked_cb(GtkWidget *button, gpointer user_data)
{
	GtkTextBuffer *code_buffer;
	gchar *c_text = NULL;
	GtkTextBuffer *asm_buffer;
	GtkTextBuffer *debug_buffer;
	GtkTextIter start;
	GtkTextIter end;
	Flow *flow;
	gchar *asm_ret = NULL;
	gchar *debug_ret = NULL;

	flow = (Flow*)user_data; /* do not free */

	/* get the C text out of the text view */
	code_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(flow->from));
	/* Obtain iters for the start and end of points of the buffer */
	gtk_text_buffer_get_start_iter (code_buffer, &start);
	gtk_text_buffer_get_end_iter (code_buffer, &end);
	/* Get the entire buffer text. */
	c_text = gtk_text_buffer_get_text (code_buffer, &start, &end, FALSE);

	compile_file(c_text, &asm_ret, &debug_ret);

	/* clear the text in the asm_view */
	asm_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(flow->to));
	gtk_text_buffer_get_start_iter (asm_buffer, &start);
	gtk_text_buffer_get_end_iter (asm_buffer, &end);
	gtk_text_buffer_delete(asm_buffer, &start, &end);

	/* if asm_ret is valid, insert it into the asm buffer */
	if (asm_ret != NULL) {
		gtk_text_buffer_get_start_iter (asm_buffer, &start);
		gtk_text_buffer_insert(asm_buffer, &start, asm_ret, -1);
	}
	
	/* if valid, append (but don't clear) the debug_ret output
		to the debug_buffer */
	if (debug_ret != NULL) {
		debug_buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(flow->debug));
		/* put the text at the end of the buffer */
		gtk_text_buffer_get_end_iter (debug_buffer, &end);
		gtk_text_buffer_insert(debug_buffer, &end, debug_ret, -1);
	}

	/* and scroll the debug window to the end of the text */
	gtk_text_buffer_get_end_iter(debug_buffer, &end);
	gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(flow->debug), 
		&end, 0.0, FALSE, 0.0, 0.0);

	/* free up anything I may have allocated */
	g_free(c_text);
	g_free(asm_ret);
	g_free(debug_ret);
}

void compile_file(gchar *c_text, gchar **asm_text, gchar **debug_output)
{
	GIOChannel *gioc = NULL;
	GError *io_err = NULL;
	GString *infile = NULL;
	GString *outfile = NULL;
	GString *gcc_cmd = NULL;
	GString *cmd = NULL;
	GString *debug = NULL;
	FILE *pin = NULL;
	gchar *asm_gcc_output = NULL;
	gsize asm_gcc_output_len;
	gsize bytes_written;
	GIOChannelError ret;
	char fbuf[1024] = {0};

	infile = g_string_new(NULL);
	outfile = g_string_new(NULL);

	/* set up the filenames I'll be using */
	g_string_printf(infile, "/tmp/c-to-asm-input-%u.c", getpid());
	g_string_printf(outfile, "/tmp/c-to-asm-output-%u.s", getpid());

	/**************************************************/
	/* write the input file */
	/**************************************************/

	/* open for writing the input file */
	gioc = g_io_channel_new_file(infile->str, "w", &io_err);
	if (gioc == NULL) {
		g_error("Couldn't open input file %s\n", infile->str);
	}

	/* write the input file */
	ret = g_io_channel_write_chars(gioc, c_text, -1, &bytes_written, &io_err);
	if (ret != G_IO_STATUS_NORMAL) {
		g_error("Couldn't write input file %s\n", infile->str);
	}

	/* were all bytes written? We're blocking, so it should have been */
	if (strlen(c_text) != bytes_written) {
		g_error("Couldn't full write input file %s\n", infile->str);
	}

	/* terminate it with a newline the input file */
	ret = g_io_channel_write_chars(gioc, "\n", -1, &bytes_written, &io_err);
	if (ret != G_IO_STATUS_NORMAL) {
		g_error("Couldn't write input file %s\n", infile->str);
	}

	/* were all bytes written? We're blocking, so it should have been */
	if (bytes_written != 1) {
		g_error("Couldn't full write newline to file %s\n", infile->str);
	}

	/* close the io channel */
	 g_io_channel_unref(gioc);

	/**************************************************/
	/* build the gcc command which will compile the process */
	/**************************************************/

	/* gcc -S infile.c -o outfile.s */
	gcc_cmd = g_string_new(NULL);
	g_string_append(gcc_cmd, "gcc -S ");
	g_string_append(gcc_cmd, infile->str);
	g_string_append(gcc_cmd, " -o ");
	g_string_append(gcc_cmd, outfile->str);

	/**************************************************/
	/* build the popen command line, which looks like:
		(echo '> ' gcc -S infile.c -o outfile.s ; 
		gcc -S infile.c -o outfile.s) 2>&1
	*/
	/**************************************************/

	cmd = g_string_new(NULL);
	g_string_append(cmd, "(echo '>' ");
	g_string_append(cmd, gcc_cmd->str);
	g_string_append(cmd, " ; ");
	g_string_append(cmd, gcc_cmd->str);
	g_string_append(cmd, ") 2>&1");

	/**************************************************/
	/* call popen, reading the output back into a buffer */
	/**************************************************/
	debug = g_string_new("");
	pin = popen(cmd->str, "r");
	if (pin == NULL) {
		g_error("Can't popen, sorry.\n");
	}
	while(fgets(fbuf, 1024, pin) != NULL)
	{
		g_string_append(debug, fbuf);
	}
	if (pclose(pin) < 0) {
		g_error("pclose failed?\n");
	}

	/**************************************************/
	/* read the asm output file, if any */
	/**************************************************/
	/* open for reading the output file */
	gioc = g_io_channel_new_file(outfile->str, "r", &io_err);
	if (gioc != NULL) {

		/* read the asm file if it exists */
		ret = g_io_channel_read_to_end(gioc, &asm_gcc_output,
										&asm_gcc_output_len, &io_err);
		if (ret != G_IO_STATUS_NORMAL) {
			g_error("Couldn't read asm file %s\n", outfile->str);
		}

		/* close the io channel */
		g_io_channel_unref(gioc);
	}

	/**************************************************/
	/* assign the return parameters for the function */
	/**************************************************/
	/* caller dellocates both of these things */
	*asm_text = asm_gcc_output;
	*debug_output = g_strdup(debug->str);

	/**************************************************/
	/* all done with the input and output files, so remove them. */
	/**************************************************/
	if (unlink(infile->str) < 0) {
		g_error("unlinking the input file broke.\n");
	}
	if (unlink(outfile->str) < 0) {
		g_error("unlinking the output file broke.\n");
	}

	/**************************************************/
	/* all done, clean up the memory */
	/**************************************************/
	g_string_free(infile, TRUE);
	g_string_free(outfile, TRUE);
	g_string_free(gcc_cmd, TRUE);
	g_string_free(cmd, TRUE);
	g_string_free(debug, TRUE);
}


int main(int argc, char **argv)
{
    GladeXML *xml;
	GtkWidget *compile_button;
	Flow *flow;

	flow = (Flow*)g_malloc(sizeof(Flow));

    gtk_init(&argc, &argv);

    xml = glade_xml_new("chasm.glade", NULL, NULL);

    /* connect signal handlers */
    glade_xml_signal_autoconnect(xml);

	/* set up the information flow that the compilation button needs to 
		worry about. */
	flow->from = glade_xml_get_widget(xml, "c_view");
	flow->to = glade_xml_get_widget(xml, "asm_view");
	flow->debug = glade_xml_get_widget(xml, "debug_view");

	/* set up the compile button callback to pass in the c_view */
	compile_button = glade_xml_get_widget(xml, "compile_button");
	g_signal_connect(compile_button,
					"clicked",
					G_CALLBACK(compile_button_clicked_cb),
					flow);

    gtk_main();

    return 0;
}
